#! /usr/bin/python

#
#  Copyright (c) Members of the EGEE Collaboration. 2004.
#  See http://eu-egee.org/partners/ for details on the copyright holders.
#  For license conditions see the license file or http://eu-egee.org/license.html
# 
#  GLite Data Transfer Scripts
#
#  Authors: Paolo Badino <paolo.badino@cern.ch>
#  Version info: $Id: transfer-monitor-gridview.py,v 1.6 2007/09/11 14:49:49 straylen Exp $
#  Release: $Name: glite-data-transfer-monitor-gridview_R_1_0_3_1 $
#
#

import os,sys,getopt,string,tempfile
import re
import fcntl
from time import gmtime, strftime, localtime, time
import cx_Oracle

# This class is needed to invoke static methods
class Callable:
    def __init__(self, anycallable):
        self.__call__ = anycallable

#
# File Locker class (only static methods)
#
class FileLocker:
    # Exclusive Lock
    LOCK_EX = fcntl.LOCK_EX
    # Shared Lock
    LOCK_SH = fcntl.LOCK_SH
    # Non Blocking
    LOCK_NB = fcntl.LOCK_NB
    
    #
    # lock a file
    #
    def lock(fh,flags):
        fcntl.flock(fh.fileno(), flags)
    lock = Callable(lock)

    #
    # unlock a file
    #
    def unlock(fh):
        try:
            fcntl.flock(fh.fileno(), fcntl.LOCK_UN)
        except IOError, detail:
            print("WARN: Failed to release lock: %s" % (detail))

    unlock = Callable(unlock)

#
# GridviewStatBuilder Class
#
class GridviewStatBuilder:
    #
    # Constructor
    #
    def __init__(self):
        self.connection    = None
        self.listCursor    = None
        self.detailCursor  = None
        self.hostRegExp    = None
        self.sdCmd         = None 
    #
    # Get Query Interval
    # 
    def queryClause(self,output_file):
        clause = None
        try:
            # Check if output file exists
            line = ''
            try:
                fh = open(output_file, 'r')
            except IOError:
                # file not found, check if the file rotated
                rotated_file = output_file + '.1'
                fh = open(rotated_file ,'r')
            # take the last line
            for line in fh:
                pass
            # Close file
            fh.close()
            if((line is not None) and (len(line) > 0)):
                # parse line
                last_fid = line.split('|')[0]
                # check that the last fid is an integer
                last_fid = int(last_fid)
                # generate clause
                clause = "fid > %d" % (last_fid)
                # print("DEBUG: last fid is %d" % last_fid)
        except ValueError, detail:
            print("WARN: Invalid file format: %s" % (detail))
            clause = None 
        except IOError, detail:
            print("WARN: Neither the output file nor the rotated one exist")
            clause = None 
        except Exception, detail:
            print("ERROR: unhandled exception in generating query clause: %s" % (detail))
            clause = None
        if(clause is None):
            # take the last 5 minutes
            clause = "logdate >= (SYSTIMESTAMP - INTERVAL '5' MINUTE)"
            print("INFO: Take all the files completed in last 5 minutes")
        return clause

    #
    # extract hostname
    #
    def extractHost(self, url):
        # Compile RegExp
        if (self.hostRegExp is None):
            self.hostRegExp = re.compile(r'^((srm|gsiftp):\/\/):?([^:\/\s]+)(.*)$')
        # find hostname
        hostname = ''
        items = self.hostRegExp.findall(url)
        if((len(items) > 0) and (len(items[0]) > 2)):
            hostname = items[0][2]
        return hostname
    #
    # get site name from hostname using service_discovery
    #
    def siteFromHost(self, hostname):
        if(len(hostname) == 0):
            return ''
        # Compile RegExp
        if (self.sdCmd is None):
            # Check if the glite-sd-query command is available in the path
            check_cmd = "which glite-sd-query > /dev/null 2>&1"
            # print("DEBUG: Executing statement <%s>" % (check_cmd))
            result = os.system(check_cmd)
            if(result == 0):
                self.sdCmd = 'glite-sd-query -t SRM  --host %s -x | grep Site'
            else:
                # Check if GLITE_LOCATION env var is set
                gl = os.environ.get('GLITE_LOCATION')
                if((gl is None) or (len(gl) == 0)):
                    # If not, go for the default location
                    gl = '/opt/glite'
                # Check if the glite-sd-query command is available in GLITE_LOCATION
                check_cmd = "which %s/bin/glite-sd-query > /dev/null 2>&1" % (gl)
                # print("DEBUG: Executing statement <%s>" % (check_cmd))
                result = os.system(check_cmd)
                
                if(result == 0):
                    os.environ["LD_LIBRARY_PATH"] = gl + '/lib' + ':/opt/globus/lib'
                    self.sdCmd = gl + '/bin/glite-sd-query -t SRM  --host %s -x | grep Site'
                else:
                    print("WARN: Cannot extract site names since no glite-sd-query command available!")
                    self.sdCmd = ''
        # find sitename
        sitename = ''
        if(len(self.sdCmd) > 0):
            try:
                cmd = self.sdCmd % (hostname)
                print("DEBUG: L_L_P %s" % os.environ.get('LD_LIBRARY_PATH') )
                print("DEBUG: GLITE_SD_PLUGIN %s " % os.environ.get('GLITE_SD_PLUGIN') )
                print("DEBUG: GFAL %s " % os.environ.get('LCG_GFAL_INFOSYS'))
                print("DEBUG: Executing %s" % cmd)
                po = os.popen(cmd, 'r')
            except IOError, detail:
                print("WARN: glite-sd-query failed: %s" % (detail))
                sitename = ''
            except Exception, detail:
                print("ERROR: unhandled exception in executing glite-sd-query : %s" % (detail))
                sitename = ''
            result = po.read()
            if(len(result) > 0):
                items = (result.splitlines()[0]).split(':')
                if(len(items) > 1):
                    sitename = items[1].strip()

        if (sitename == ''):
            sitename = 'UNKNOWN'

        return sitename

    #
    # connect
    #
    def connect(self, u,p,c):
        try:
            self.connection   = cx_Oracle.connect(u,p,c)
            self.listCursor   = self.connection.cursor()
            self.detailCursor = self.connection.cursor()
        except cx_Oracle.DatabaseError,detail:
            print "ERROR: No Connection Established to Database \nERROR: %s %s" % (cx_Oracle.DatabaseError,detail)
            return -1
        return 0

    # validate 
    #
    def validate(self):
        try:
            # Create a Query that check the DB schema is correct
            query = 'SELECT major, minor, patch FROM t_schema_vers WHERE ROWNUM <= 1' 
            self.detailCursor.execute(query);
            row = self.detailCursor.fetchone()
            major = int(row[0])
            minor = int(row[1])
            patch = int(row[2])
            if((major == 2 and minor >= 2) or (major == 3)):
                # Schema is ok. Now check that the t_filestats table is there
                query = 'SELECT fid FROM t_filestats WHERE ROWNUM <= 1' 
                try:
                    self.detailCursor.execute(query);
                    row = self.detailCursor.fetchone()
                    # print("Validation succeeded. DB Schema is %d.%d.%d" % (major,minor,patch))
                except cx_Oracle.DatabaseError,detail:
                    print("ERROR: Validation failed. Filestats Log table missing")
                    print("ERROR: %s" % (detail))
                    print("ERROR: Please be sure that the fts_stats_pack is installed and the table is visible to the account in use")
                    return -1
            else:
                print("ERROR: Unsupported schema. Current number is %d.%d.%d" % (major,minor,patch))
                return -1
        except ValueError, detail:
            print("ERROR: Invalid Schema Number \nERROR: %s" % (detail))
            return -1
        except cx_Oracle.DatabaseError,detail:
            print("ERROR: No Connection Established to Database \nERROR: %s %s" % (cx_Oracle.DatabaseError,detail))
            return -1
        except Exception, detail:
            print("ERROR: unhandled exception in validation: %s" % (detail))
            return -1
      
        # return success
        return 0
    #
    # generate 
    #
    def generate(self,output_file):
        # Create Queries
        list_query    = None
        detail_query  = None
        clause = self.queryClause(output_file)
        list_query = '''
SELECT * FROM (
SELECT L.fid, L.logdate AT TIME ZONE '+00:00', L.job_id, L.file_id, L.vo_name, L.channel_name, L.filesize, L.finish_time
FROM t_filestats L 
WHERE (%CLAUSE%)
ORDER BY L.fid ASC
) WHERE ROWNUM <= 5000
'''
        # Replace placeholders
        list_query = list_query.replace('%CLAUSE%'   , clause)
        detail_query = '''
SELECT F.source_surl,F.dest_surl
FROM t_file F
WHERE F.file_id = :1
'''
        # print("DEBUG: List query \n%s\n" % list_query)
        # print("DEBUG: Detail query \n%s\n" % destil_query)
        try:
            try:
                # open output file for writing
                fh = open(output_file,'a')
                try:
                    FileLocker.lock(fh,FileLocker.LOCK_EX)
                except IOError,detail:
                    print("ERROR: Failed to lock output file: %s" % (detail) )
                    return -1

                # Process Query and generate Report
                self.listCursor.execute(list_query)   
                rows = self.listCursor.fetchall()
                #print("DEBUG: Got %d row(s)" % (len(rows)) )

                # Get rows
                for row in rows:
                    fid          = row[0]
                    logdate      = row[1]
                    job_id       = row[2]
                    file_id      = row[3]
                    vo_name      = row[4]
                    channel_name = row[5]
                    file_size    = row[6]
                    finish_time  = row[7]
                    # postprocess the file_size
                    if((file_size is None) or (len(str(file_size)) == 0) or (str(file_size) == '0')):
                        file_size = -1
                    # postprocess the finish_time
                    if((finish_time is None) or (len(str(finish_time)) == 0) or (str(finish_time) == '0')):
                        finish_time = -1
                    source_file  = ''
                    dest_file    = ''
                    source_host  = ''
                    dest_host    = ''
                    source_site  = ''
                    dest_site    = ''
                    # Process Query and generate Report
                    params = {'1':file_id}
                    self.detailCursor.execute(detail_query,params)
                    details = self.detailCursor.fetchall()
                    if(len(details) > 0):
                        source_file  = details[0][0]
                        dest_file    = details[0][1]
                        # parse SURLs and extract host name
                        source_host = self.extractHost(source_file)
                        dest_host   = self.extractHost(dest_file)
                        # use ServiceDiscovery to get the site name
                        source_site = self.siteFromHost(source_host)
                        dest_site   = self.siteFromHost(dest_host)

                    fh.write("%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s\n" % (fid,logdate,job_id,file_id,vo_name,channel_name,file_size,finish_time, source_file,dest_file,source_host,dest_host,source_site,dest_site))

            except cx_Oracle.DatabaseError,detail:
                print("ERROR: DB error: %s %s" % (cx_Oracle.DatabaseError,detail))
                print("DEBUG: Queries are:\n[%s]\nand\n[%s]\nERROR: %s %s" % (list_query,detail_query))
                return -1
            except IOError,detail:
                print("ERROR: Cannot write output file: %s" % (detail) )
                return -1
            except Exception, detail:
                print("ERROR: unhandled exception in generating report: %s" % (detail))
                return -1
      
        finally:
            # Release lock
            FileLocker.unlock(fh)
            # Close file
            fh.close()
        # return success
        return 0
 
    #
    # generateSummary 
    #
    def generateSummary(self,type,output_file):
        # Create Queries
        list_query    = None

        if(type == 'hour'):
            list_query = '''
SELECT 'H', TO_TIMESTAMP_TZ(TO_CHAR(SYSTIMESTAMP AT TIME ZONE '+00:00','YYYY-MM-DD HH24') || ' :00:00 +00:00','YYYY-MM-DD HH24:MI:SS TZH:TZM')
     , MIN(L.fid), MAX(L.fid)
     , MIN(L.logdate) AT TIME ZONE '+00:00', MAX(L.logdate) AT TIME ZONE '+00:00'
     , COUNT(L.FID)
FROM t_filestats L
where L.logdate >= TO_TIMESTAMP_TZ((TO_CHAR(SYSTIMESTAMP - INTERVAL '1' HOUR,'YYYY-MM-DD HH24') || ' :00:00 ' || TO_CHAR(TZ_OFFSET(SESSIONTIMEZONE))) ,'YYYY-MM-DD HH24:MI:SS TZH:TZM')
and L.logdate < TO_TIMESTAMP_TZ((TO_CHAR(SYSTIMESTAMP,'YYYY-MM-DD HH24') || ' :00:00 ' || TO_CHAR(TZ_OFFSET(SESSIONTIMEZONE))) ,'YYYY-MM-DD HH24:MI:SS TZH:TZM')
'''
        elif(type == 'day'):
            list_query = '''
SELECT 'D', TO_TIMESTAMP_TZ(TO_CHAR(SYSTIMESTAMP AT TIME ZONE '+00:00','YYYY-MM-DD HH24') || ' :00:00 +00:00','YYYY-MM-DD HH24:MI:SS TZH:TZM')
     , MIN(L.fid), MAX(L.fid)
     , MIN(L.logdate) AT TIME ZONE '+00:00', MAX(L.logdate) AT TIME ZONE '+00:00'
     , COUNT(L.FID)
FROM t_filestats L
where L.logdate >= TO_TIMESTAMP_TZ((TO_CHAR(SYSDATE  - INTERVAL '1' DAY,'YYYY-MM-DD') || ' 00:00:00 ' || TO_CHAR(TZ_OFFSET(SESSIONTIMEZONE))) ,'YYYY-MM-DD HH24:MI:SS TZH:TZM')
and L.logdate < TO_TIMESTAMP_TZ((TO_CHAR(SYSDATE,'YYYY-MM-DD') || ' 00:00:00 ' || TO_CHAR(TZ_OFFSET(SESSIONTIMEZONE))) ,'YYYY-MM-DD HH24:MI:SS TZH:TZM')
'''
        elif(type == 'week'):
            list_query = '''
SELECT 'W', TO_TIMESTAMP_TZ(TO_CHAR(SYSTIMESTAMP AT TIME ZONE '+00:00','YYYY-MM-DD HH24') || ' :00:00 +00:00','YYYY-MM-DD HH24:MI:SS TZH:TZM')
     , MIN(L.fid), MAX(L.fid)
     , MIN(L.logdate) AT TIME ZONE '+00:00', MAX(L.logdate) AT TIME ZONE '+00:00'
     , COUNT(L.FID)
FROM t_filestats L
where L.logdate >= TO_TIMESTAMP_TZ((TO_CHAR(SYSDATE  - INTERVAL '7' DAY,'YYYY-MM-DD') || ' 00:00:00 ' || TO_CHAR(TZ_OFFSET(SESSIONTIMEZONE))) ,'YYYY-MM-DD HH24:MI:SS TZH:TZM')
and L.logdate < TO_TIMESTAMP_TZ((TO_CHAR(SYSDATE,'YYYY-MM-DD') || ' 00:00:00 ' || TO_CHAR(TZ_OFFSET(SESSIONTIMEZONE))) ,'YYYY-MM-DD HH24:MI:SS TZH:TZM')
'''
        else:
            print("ERROR: Unsupported summary type %s. Allowed values are [hour,day,week]" % (type))
            return -1
         
        # print("DEBUG: List query \n%s\n" % list_query)
        try:
            try:
                # open output file for writing
                fh = open(output_file,'a')
                try: 
                    FileLocker.lock(fh,FileLocker.LOCK_EX)
                except IOError,detail:
                    print("ERROR: Failed to lock output file: %s" % (detail) )
                    return -1

                # Process Query and generate Report
                self.listCursor.execute(list_query)   
                rows = self.listCursor.fetchall()

                # Get rows
                for row in rows:
                    t            = row[0]
                    ts           = row[1]
                    min_fid      = row[2]
                    max_fid      = row[3]
                    min_logdate  = row[4]
                    max_logdate  = row[5]
                    count        = row[6]
                    fh.write("%s|%s|%s|%s|%s|%s|%s\n" % (t,ts,min_fid,max_fid,min_logdate,max_logdate,count))

            except cx_Oracle.DatabaseError,detail:
                print("ERROR: DB error: %s %s" % (cx_Oracle.DatabaseError,detail))
                print("DEBUG: Queries are:\n[%s]\nand\n[%s]\nERROR: %s %s" % (list_query,detail_query))
                return -1
            except IOError,detail:
                print("ERROR: Cannot write output file: %s" % (detail) )
                return -1
            except Exception, detail:
                print("ERROR: unhandled exception in generating report: %s" % (detail))
                return -1
      
        finally:
            # Release lock
            FileLocker.unlock(fh)
            # Close file
            fh.close()
 
        # return success
        return 0

    #
    # disconnect
    #
    def disconnect(self):
        self.detailCursor.close()
        self.listCursor.close()
        self.connection.close()

###############################################################################

#
# Print usage
#
def help_message(str = ''):
    if str != '':
        print '\nERROR: %s' % (str)
        
    print '''
Syntax:
    %s -u <name> -p <password> -c <str> [-S <type>] [-o output]
Options:
    -h     -- display this help
    -u     -- Database User's Name
    -p     -- Database User's Password
    -c     -- Database Connection string
    -o     -- The name of the file where the data has to be generated (default: output.dat)
    -S     -- Generate a summary in order to verify the consistency of the data. The value of <type> gives the scope of the summary
''' % sys.argv[0]
    
    sys.exit(0)

#
# acquire process lock
#
def single_instance_lock(lock_file):
    try:
        fh = open(lock_file, "w")
        #if (0 != FileLocker.lock(fh,FileLocker.LOCK_EX | FileLocker.LOCK_NB)):
        try:
            FileLocker.lock(fh,FileLocker.LOCK_EX | FileLocker.LOCK_NB)
        except IOError, detail:
            print("ERROR: Cannot acquire lock, probably an instance is already running: %s" % (detail))
            sys.exit(-1)
        # write process PID into the lock file
        fh.write(str(os.getpid()))
    except IOError, detail:
        print("ERROR: Cannot write lock file: %s" % (detail) )
        sys.exit(-1)
    except Exception, detail:
        print("ERROR: unhandled exception in acquiring process lock: %s" % (detail))
        sys.exit(-1)
   
    return fh

#
# release process lock
#
def single_instance_unlock(lock_file, fh):
    try:
        # unlock file 
        FileLocker.unlock(fh)
        #print("DEBUG: file lock released")
        # Close file
        fh.close()
    finally:
        # Remove lock file
        os.unlink(lock_file)

#
# Main
#
def main():

    try:
        opts, args = getopt.getopt(sys.argv[1:],'hu:p:c:o:S:Z')
    except getopt.error, msg:
        help_message(msg)

    result         = 0
    # default values
    user_name      = ''
    password       = ''
    connect_str    = ''
    output_file    = ''
    lock_file      = ''
    validation_mode= False
    summary_type   = ''
    # Get the arguments
    for o,a in opts[:]:
        if(o == '-u'):
            user_name = a
        elif(o == '-p'):
            password = a
        elif(o == '-c'):
            connect_str = a
        elif(o == '-o'):
            output_file = a
        elif(o == '-S'):
            summary_type = a
        elif(o == '-Z'):
            validation_mode = True
        elif(o == '-h'):
            help_message()

    # Check output file name
    if(len(output_file) == 0):
        if(len(summary_type) > 0):
            output_file = 'summary.dat' 
        else:
            output_file = 'output.dat'
    # Check lock file name
    if(len(lock_file) == 0):
        if(len(summary_type) > 0):
            lock_file = output_file + "." + summary_type + ".lock" 
        else:
            lock_file = output_file + ".lock" 
    # lock
    fh = single_instance_lock(lock_file) 
    try:
        # Check the parameters
        if(len(user_name) == 0):
            help_message('No Username specified')
        if(len(password) == 0):
            help_message('No Password specified')
        if(len(connect_str) == 0):
            help_message('No DB Connection string specified')
        # create stat generator instance
        r = GridviewStatBuilder()
        # Create connection to the DB and create output file
        result = r.connect(user_name,password,connect_str)
        if(result == 0):
            if(validation_mode == True):
                # validate
                result = r.validate()
            elif(len(summary_type) > 0):
                # Generate Summary Report
                result = r.generateSummary(summary_type,output_file)
            else:
                # Generate Report
                result = r.generate(output_file)
        # Close the connection to the DB and close output file
        r.disconnect()
        
    finally:
        # unlock
        single_instance_unlock(lock_file,fh) 
    # Exit
    sys.exit(result)

if __name__ == "__main__":
    main()

